%:-dynamic(c/1).
:-dynamic(eq/3).
:-dynamic(ruleold/3).
:-dynamic(rulecur/3).
:-dynamic(ruleon/4).
:-dynamic(rule/3).
:-dynamic(grprec/2).
:-dynamic(eqprec/2).
:-dynamic(status/2).

%eq(left,right,red) --> red is 0/1  
%eqold(left,right,red)

complete:-
    initialize,
    write('File: '),
    read(F),
    makefilename(F,NF),
    see(NF),
    repeat,
    read(T),
    treatall(T),!,
    seen,
%    nl,nl,
%    show_order,!,
    (eq(_,_,_) ->
     (nl,nl,write('Initial set of equations: '),nl,nl,
    show_equations,!,nl,
    orientEquations,
%    tell(result),
    nl,nl,write('Initial set after orienting and simplifying: '),nl,nl,
    show_rules,!,nl,nl,
    completion)
    ; true),
    told
    ,!.

makefilename(Name,Name):- 
    file_exists(Name).
makefilename(Name,FName):- 
    name(Name,Lname),
    name('.com',Lext),
    append(Lname,Lext,Lfile),
    name(FName,Lfile),
    file_exists(FName).
makefilename(_,_):- 
    write('Error: File does not exist'),nl,nl,!,fail.


initialize:-
    told,
    seen,
    clean_all,
    g_assign(freshmax,0),
    g_assign(freshcounter,0).
%    ,
%        op(600, xfy, ->),
%        op(600, xfy, =),
%        op(600, xfy, >),
%        g_assign(noproof,0),
%        g_assign(typecons,[]),

treatall(end_of_file).

treatall(T):-
    treat(T), fail.

treat(eq(C,T1,T2)):-!,
      assertz(eq(C,T1,T2)).
treat(grF(S1,S2)):-!,
      assertz(grprec(S1,S2)).
treat(eqF(S1,S2)):-!,
      assertz(eqprec(S1,S2)).
treat(stat(S,St)):-!,
      member(S,[mul,lex-lr,lex-rl]),
      assertz(status(S,St)).

clean_all:-
    retractall(rule(_,_,_)),
    retractall(eq(_,_,_)),
    retractall(ruleold(_,_,_)),
    retractall(rulecur(_,_,_)),
    retractall(ruleon(_,_,_)),
    retractall(grprec(_,_)),
    retractall(eqprec(_,_)),
    retractall(status(_,_)),
    !.

orientEquations:-
    retract(eq(C,T1,T2)),
    simplify(C,T1,T2,NT1,NT2),
    checkresult(C,T1,T2,NT1,NT2),
    compare(C,NT1,NT2,T11,T21,RED),
    (RED=0 -> (write('unorientable'),nl,!,abort)
      ; 
     true
    ),
    refresh(C,T11,_,CF,T11F,_),
    assertz(rulecur(C,T11,T21)),
    simplifyback(CF,T11F),
    retract(rulecur(_,_,_)),
    assertz(rule(C,T11,T21)),
    fail.
orientEquations.

checkresult(C,S1,S2,NS1,NS2):-
	tautology(C,NS1,NS2),
	write('    The equation  '),nl,
	write('     '), write_equation(C,S1,S2),nl,
	write('    is redundant'),nl,!,fail.
checkresult(C,S1,S2,NS1,NS2):-
	(S1\==NS1 ; S2\==NS2),
	write('    The equation  '),nl,
	write('     '), write_equation(C,S1,S2),nl,
	write('    has been simplified into  '),nl,
	write('     '), write_equation(C,NS1,NS2),nl,!.
checkresult(_,_,_,_,_).


compare(C,T1,T2,T1,T2,1):-
    greater_ord(C,T1,T2),!.
compare(C,T1,T2,T2,T1,1):-
    greater_ord(C,T2,T1),!.
compare(_C,T1,T2,T1,T2,0):-!.

checkDifAlpha([],_).
checkDifAlpha([(A,_)|EC],C):-
    isFreshMatch(A),!,
    checkDifAlpha(EC,C),!.
checkDifAlpha([(A,X)|EC],C):-
    member1((A,X),C),!,
    checkDifAlpha(EC,C),!.

tautology(C,S,T):-
    tautology([],S,[],T,UC),
    simplifyContext(UC,SC),
    checkDifAlpha(SC,C).

tautology(PermS,S,PermT,T,UC):-var(S),var(T),!,S==T,
    difPi(PermS,PermT,S,UC).
tautology(PermS,S,PermT,perm(LP,T),UC):-var(S),!,
    append(PermT,LP,PermT1),
    tautology(PermS,S,PermT1,T,UC),!.
tautology(_,S,_,_,_):-var(S),!,fail.
tautology(PermS,perm(LP,S),PermT,T,UC):-var(T),!,
    append(PermS,LP,PermS1),
    tautology(PermS1,S,PermT,T,UC),!.
tautology(_,_,_,T,_):-var(T),!,fail.
tautology(PermS,S,PermT,T,[]):-
    S=..[A], 
    T=..[B],!,
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,A1),!.
tautology(PermS,abs(A,S),PermT,abs(B,T),UC1):-
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,B1),
    (A1==B1 -> 
     (tautology(PermS,S,PermT,T,UC),
      UC1=UC)
     ; 
     (tautology(PermS,S,[(A1,B1)|PermT],T,UC),
      UC1=[(A1,T)|UC])),!.
tautology(PermS,perm(LP,S),PermT,T,UC):-!,
    append(PermS,LP,PermS1),
    tautology(PermS1,S,PermT,T,UC),!.
tautology(PermS,S,PermT,perm(LP,T),UC):-!,
    append(PermT,LP,PermT1),
    tautology(PermS,S,PermT1,T,UC),!.
tautology(PermS,S,PermT,T,UC):-
    S=..[F|ArgsS], 
    T=..[F|ArgsT],!,
    tautologylist(PermS,ArgsS,PermT,ArgsT,UC),!.

tautologylist(_,[],_,[],[]).
tautologylist(PermS,[S1|ArgsS],PermT,[T1|ArgsT],UC):-
    tautology(PermS,S1,PermT,T1,UC1),!,
    tautologylist(PermS,ArgsS,PermT,ArgsT,UC2),
    append(UC1,UC2,UC),!.

completion:-
    repeat,
    completion1,
    chk_finish.


completion1:-
    retract(rule(C,T1,T2)),
    assertz(ruleold(C,T1,T2)),
    infer(C,T1,T2),!.

chk_finish:-
    rule(_,_,_),!,
    fail.
chk_finish:-
    nl,nl,write('Final set of equations: '),nl,
    show_equations.

show_order:-
	write('The nrpo is defined from the transitive closure of the given precedence:'), nl,nl,
%	show_order1,nl,
	!.

show_equations:-
	g_assign(eqnum,0),
	show_equations1.

show_rules:-
	g_assign(rnum,0),
	show_rules1.

write_rule(C,E1,E2):- 
    \+ ( \+ (write_rule1(C,E1,E2))).
write_equation(C,E1,E2):- 
    \+ ( \+ (write_equation1(C,E1,E2))).

write_rule(C,E1,E2,N):- 
    write(N), write(') '),
    \+ ( \+ (write_rule1(C,E1,E2))).
write_equation(C,E1,E2,N):- 
    write(N), write(') '),
    \+ ( \+ (write_equation1(C,E1,E2))).

write_rule1(C,E1,E2):-
    write('['),
    write_context(C,0,C1), 
    write('] |-- '),
    write_term1(E1,C1,C2),write(' --> '),write_term1(E2,C2,_),!.

write_equation1(C,E1,E2):-
    write('['),
    write_context(C,0,C1), 
    write('] |-- '),
    write_term1(E1,C1,C2),write(' == '),write_term1(E2,C2,_),!.

write_term1(X,N,N1):-
    var(X),
    N1 is N+1,
    name(N1,L), 
    name(X,[88|L]), %x1,x2,x3,.....
    write(X),!.
write_term1([A|R],N,N1):-
    write('['),
    write_term1_list([A|R],N,N1),
    write(']'),!.
write_term1((A,B),N,N):-
    write('('), write(A), write(','),
                write(B), write(')'),!.
write_term1(T,N,N1):-
    T=..[F|A],
    write(F),
    (
     (A=[],N1=N)
     ;
     (
      write('('),
      write_term1_list(A,N,N1),
      write(')')
     )
    ),!.

write_term1_list([],N,N):-!.
write_term1_list([T],N,N1):-!,
	write_term1(T,N,N1),!.
write_term1_list([T|A],N,N2):-
	write_term1(T,N,N1),!,
	write(','),
	write_term1_list(A,N1,N2),!.

write_context([],N,N):-!.
write_context([P],N,N1):-!,
    write_context_pair(P,N,N1),!.
write_context([P|C],N,N2):-
    write_context_pair(P,N,N1),!,
    write(','),
    write_context(C,N1,N2),!.
    
write_context_pair((A,T),N,N1):-
        write('('),
	write(A),
        write(','),
	write_term1(T,N,N1),
        write(')'),!.
	
show_equations1:-
    eq(C,E1,E2),
    g_read(eqnum,N),
    N1 is N+1,
    write_equation(C,E1,E2,N1),nl,
    g_assign(eqnum,N1),
    fail.
show_equations1:-
    ruleold(C,E1,E2),
    g_read(eqnum,N),
    N1 is N+1,
    write_rule(C,E1,E2,N1),nl,
    g_assign(eqnum,N1),
    fail.
show_equations1:-
    rule(C,E1,E2),
    g_read(eqnum,N),
    N1 is N+1,
    write_rule(C,E1,E2,N1),nl,
    g_assign(eqnum,N1),
    fail.
show_equations1.

show_rules1:-
    rule(C,E1,E2),
    g_read(rnum,N),
    N1 is N+1,
    write_rule(C,E1,E2,N1),nl,
    g_assign(rnum,N1),
    fail.
show_rules1.


infer(C,T1,T2):-
    nl,write('Inferences with: '), write_rule(C,T1,T2),nl,
    %read(_),
    inferold(C,T1,T2),
    infernew(C,T1,T2),!.

inferold(C1,T1,T2):-
    ruleold(C2o,T3o,T4o),
    nl,write('    Inferences on old: '), write_rule(C2o,T3o,T4o),nl,
    refresh(C2o,T3o,T4o,C2,T3,T4),
    %nl,write('    refreshed: '), write_rule(C2,T3,T4),nl,
    infernotop(C1,T1,T2,C2,T3,T4),
    fail.
inferold(_,_,_).

infernew(C1,T1,T2):-
    rule(C2o,T3o,T4o),
    nl,write('    Inferences on new: '), write_rule(C2o,T3o,T4o),nl,
    refresh(C2o,T3o,T4o,C2,T3,T4),
    %nl,write('    refreshed: '), write_rule(C2,T3,T4),nl,
    infer(C1,T1,T2,C2,T3,T4),
    fail.
infernew(_,_,_).

infer(C1,T1,T2,C2,T3,T4):-
    subtnovar(T3,S,C,X),
    assertz(ruleon(C2,T3,T4)),
    inferone(C1,T1,T2,S,C,X,C2,T4),
    retract(ruleon(_,_,_)),
    fail.
infer(_,_,_,_,_,_):-!.

infernotop(C1,T1,T2,C2,T3,T4):-
    subtnovar(T3,S,C,X),C\==X,
    assertz(ruleon(C2,T3,T4)),
    inferone(C1,T1,T2,S,C,X,C2,T4),
    retract(ruleon(_,_,_)),
    fail.
infernotop(_,_,_,_,_,_):-!.

difPi(PiX,PiY,X,UC):-
    getAtoms([],PiX,LX),
    getAtoms(LX,PiY,LXY),    
    getDifs(LXY,PiX,PiY,X,UC),!.

getDifs([],_,_,_,[]).
getDifs([A|L],PiX,PiY,X,UC):-
    applyPermAtom(PiX,A,C1),
    applyPermAtom(PiY,A,C2),
    C1==C2,!,
    getDifs(L,PiX,PiY,X,UC).
getDifs([A|L],PiX,PiY,X,[(A,X)|UC]):-
    getDifs(L,PiX,PiY,X,UC).

getAtoms(L,[],L).
getAtoms(L,[(A,B)|R],L3):-
    addAtom(A,L,L1),
    addAtom(B,L1,L2),
    getAtoms(L2,R,L3).

addAtom(A,[],[A]).
addAtom(A,[A|L],L):-!.
addAtom(A,[B|L],[B|L1]):-
    addAtom(A,L,L1).


unifyabs(T1,T2,UC):-unifyabs([],T1,[],T2,UC).

% unifyabs(CS,PermS,S,CT,PermT,T,UC) unifies S and T under the permutations
% PermS and PermT (applying the unifier) and then context CS and CT
% and returns the unification context UC

unifyabs(PermS,S,PermT,T,UC):-var(S),var(T),S==T,!,
    difPi(PermS,PermT,S,UC).
unifyabs(PermS,S,PermT,T,[]):-var(S),var(T),!,
    applyPermut(PermT,T,T1),
    reverse(PermS,RPermS),
    applyPermut(RPermS,T1,T2),
    S=T2,!.
unifyabs(PermS,S,PermT,perm(LP,T),UC):-var(S),!,
    append(PermT,LP,PermT1),
    unifyabs(PermS,S,PermT1,T,UC),!.
unifyabs(PermS,S,PermT,T,[]):-var(S),!,
    \+issubterm(S,T),
    applyPermut(PermT,T,T1),
    reverse(PermS,RPermS),
    applyPermut(RPermS,T1,T2),
    S=T2,!.
unifyabs(PermS,perm(LP,S),PermT,T,UC):-var(T),!,
    append(PermS,LP,PermS1),
    unifyabs(PermS1,S,PermT,T,UC),!.
unifyabs(PermS,S,PermT,T,[]):-var(T),!,
    \+issubterm(T,S),
    applyPermut(PermS,S,S1),
    reverse(PermT,RPermT),
    applyPermut(RPermT,S1,S2),
    T=S2,!.
unifyabs(PermS,S,PermT,T,[]):-
    S=..[A], 
    T=..[B],!,
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,A1),!.
unifyabs(PermS,abs(A,S),PermT,abs(B,T),UC1):-
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,B1),
    (A1==B1 -> 
     (unifyabs(PermS,S,PermT,T,UC),
      UC1=UC)
     ; 
     (unifyabs(PermS,S,[(A1,B1)|PermT],T,UC),
      UC1=[(A1,T)|UC])),!.
unifyabs(PermS,perm(LP,S),PermT,T,UC):-!,
    append(PermS,LP,PermS1),
    unifyabs(PermS1,S,PermT,T,UC),!.
unifyabs(PermS,S,PermT,perm(LP,T),UC):-!,
    append(PermT,LP,PermT1),
    unifyabs(PermS,S,PermT1,T,UC),!.
unifyabs(PermS,S,PermT,T,UC):-
    S=..[F|ArgsS], 
    T=..[F|ArgsT],!,
    unifyabslist(PermS,ArgsS,PermT,ArgsT,UC),!.

unifyabslist(_,[],_,[],[]).
unifyabslist(PermS,[S1|ArgsS],PermT,[T1|ArgsT],UC):-
    unifyabs(PermS,S1,PermT,T1,UC1),!,
    unifyabslist(PermS,ArgsS,PermT,ArgsT,UC2),
    append(UC1,UC2,UC),!.

applyPermut([],T,T):-!.
applyPermut(Perm,T,perm(Perm,T)):-
    var(T),!.
applyPermut(Perm,T,B):-
    T=..[A],!, 
    applyPermAtom(Perm,A,B),!.
applyPermut(Perm,abs(A,T),abs(B,T1)):-
    applyPermAtom(Perm,A,B),
    applyPermut(Perm,T,T1),!.
applyPermut(Perm,perm(LP,T),T1):-!,
    append(Perm,LP,Perm1),
    applyPermut(Perm1,T,T1).
applyPermut(Perm,T,T1):-
    T=..[F|Args],!, 
    applyPermutList(Perm,Args,Args1),
    T1=..[F|Args1],!.

applyPermutList(_,[],[]).
applyPermutList(Perm,[T|Args],[T1|Args1]):-
    applyPermut(Perm,T,T1),
    applyPermutList(Perm,Args,Args1),!.

applyPermAtom([],A,A).
applyPermAtom([(C,B)|Perm],A,A1):-
    applyPermAtom(Perm,A,A2),
    ((A2==C,A1=B);
     (A2==B,A1=C);
     A1=A2),!.

simplifyContext(C,CS):-
    simplifyContext1(C,C1),!,
    delRepPairs(C1,CS).

simplifyContext1([],[]).
simplifyContext1([(A,T)|R],L):-
    simplifyCondition(A,T,L1),
    simplifyContext1(R,LR),
    append(L1,LR,L).

delRepPairs([],[]).
delRepPairs([P|C1],[P|CS]):-
    deleteall(P,C1,C2),!,
    delRepPairs(C2,CS).

simplifyCondition(A,T,[(A,T)]):-var(T),!.
simplifyCondition(A,A,_):-!,fail.
simplifyCondition(A,abs(A,_),[]).
simplifyCondition(A,perm(L,T),UC):-
    reverse(L,L1),
    applyPermAtom(L1,A,B),
    simplifyCondition(B,T,UC).
simplifyCondition(A,T,L):-
    T=..[_|Args],!,
    simplifyConditionlist(A,Args,L).

simplifyConditionlist(_,[],[]).
simplifyConditionlist(A,[T1|Args],L):-
    simplifyCondition(A,T1,L1),
    simplifyConditionlist(A,Args,L2),
    append(L1,L2,L).

simplifyPerms(X,X):-var(X),!.
simplifyPerms(perm(L,X),T):-
    reducePerm(L,L1),
    (L1=[] -> T=X
     ; T= perm(L1,X)
    ),!.
simplifyPerms(abs(A,T),abs(A,T1)):-!,
    simplifyPerms(T,T1).
simplifyPerms(T,T1):-!,
    T=..[F|Args],
    simplifyPermsList(Args,Args1),
    T1=..[F|Args1],!.

reducePerm([],[]).
reducePerm([(A,B)|L],LF):-
    reducePerm(L,L1),
    (deletePerm((A,B),L1,LF);
     LF=[(A,B)|L1]),!.

deletePerm((A,B),[(A,B)|L],L):-!.
deletePerm((A,B),[(B,A)|L],L):-!.
deletePerm((A,B),[(C,D)|_],_):-
    (A=C;B=C;A=D;B=D),!,fail.
deletePerm((A,B),[(C,D)|L],[(C,D)|L1]):-!,
    deletePerm((A,B),L,L1),!.

simplifyPermsList([],[]).
simplifyPermsList([T|Args],[T1|Args1]):-
    simplifyPerms(T,T1),
    simplifyPermsList(Args,Args1),!.

deleteall(_,[],[]).
deleteall(P,[P1|C1],C2):-
    P==P1,!,
    deleteall(P,C1,C2).
deleteall(P,[P1|C1],[P1|C2]):-
    deleteall(P,C1,C2).

downPerm(X,X):-var(X).
downPerm(perm(P,T),TF):-
    downPerm(T,T1),
    applyPermut(P,T1,TF).
downPerm(T,TF):-
    T=..[F|Args],
    downPermList(Args,ArgsF),
    TF=..[F|ArgsF],!.

downPermList([],[]).
downPermList([T|Args],[TF|ArgsF]):-
    downPerm(T,TF),
    downPermList(Args,ArgsF),!.

inferone(C1,T1,T2,S,C,X,C2,T4):-
    unifyabs(T1,S,RC),
    X=T2,
    append(RC,C1,CR1),
    append(CR1,C2,C12),
    simplifyContext(C12,C12s),
    downPerm(C,Cp),
    downPerm(T4,T4p),
    %ruleon(CO,O1,O2),
    %write('    on: '),
    %write('     '), write_rule(CO,O1,O2),nl,
    write('    New equation obtained: '),nl,
    write('     '), write_equation(C12s,Cp,T4p),nl,
    simplify(C12s,Cp,T4p,NT1,NT2),
    % write('    After simplification: '),nl,
    % write('     '), write_equation(C12s,NT1,NT2),nl,
    checkresult(C12s,Cp,T4p,NT1,NT2),
    compare(C12s,NT1,NT2,T11,T21,RED),
    (RED=0 -> (write('    '), write('unorientable'),nl,!,abort)
      ; 
     (	simplifyContextTerm(C12s,pair(T11,T21),CF),
	refresh0(CF,T11,T21,CF0,T10,T20),
	write('   The rule  '),nl,
	write('    '), write_rule(CF0,T10,T20),nl,
	write('   has been added'),nl,
	assertz(rulecur(CF0,T10,T20)),
	simplifyback(CF,T11),
	retract(rulecur(_,_,_)),
	assertz(rule(CF0,T10,T20)) )
    ),
%    read(_),
    !.    
inferone(_,_,_,_,_,_,_,_).

simplifyContextTerm(C1,T,C2):-
    getvarsandnames(T,[],L),
    simplifyContextTerm1(C1,L,C2),!.

getvarsandnames(X,L,[X|L]):-var(X),!.
getvarsandnames(abs(A,T),L,[A|L1]):-!,
    getvarsandnames(T,L,L1),!.
getvarsandnames(T,L,L1):-
    T=..[_|Args],
    getvarsandnameslist(Args,L,L1),!.

getvarsandnameslist([],L,L):-!.
getvarsandnameslist([T|Args],L,L2):-
    getvarsandnames(T,L,L1),
    getvarsandnameslist(Args,L1,L2),!.

simplifyContextTerm1([],_,[]).
simplifyContextTerm1([(A,V)|C1],L,[(A,V)|C2]):-
    member1(A,L),
    member1(V,L),
    simplifyContextTerm1(C1,L,C2),!.
simplifyContextTerm1([_|C1],L,C2):-
    simplifyContextTerm1(C1,L,C2),!.

%Y is a variable
checkVar(C,X,PiX,Y,PiY):-
    X==Y,!,
    difPi(PiX,PiY,X,EC),
    checkDif(EC,PiX,PiY,C),!.
checkVar(_,X,_,_,_):-
    var(X),!,fail.
checkVar(C,perm(LT,T),PiT,Y,PiY):-!,
    append(PiT,LT,EPiT),
    checkVar(C,T,EPiT,Y,PiY).
checkVar(C,abs(A,T),PiT,Y,PiY):-
    applyPermAtom(PiT,A,A1),
    freshordname(B),
    checkVar(C,T,[(A1,B)|PiT],Y,PiY),!.
checkVar(C,T,PiT,Y,PiY):-
    T=..[_|Args],
    checkVarList(C,Args,PiT,Y,PiY),!.

checkVarList(C,[T|_],PiT,Y,PiY):-
    checkVar(C,T,PiT,Y,PiY),!.
checkVarList(C,[_|Args],PiT,Y,PiY):-
    checkVarList(C,Args,PiT,Y,PiY).

checkDif([],_,_,_).
checkDif([(A,_)|EC],PiX,PiY,C):-
    isFreshOrd(A),!,
    checkDif(EC,PiX,PiY,C),!.
checkDif([(A,_)|EC],PiX,PiY,C):-
    applyPermAtom(PiX,A,A1),
    isFreshOrd(A1),
    applyPermAtom(PiY,A,A2),
    isFreshOrd(A2),!,
    checkDif(EC,PiX,PiY,C),!.
checkDif([(A,X)|EC],PiX,PiY,C):-
    member1((A,X),C),!,
    checkDif(EC,PiX,PiY,C),!.

greater_ord(C,S,T):-
    g_assign(freshordcounter,0),
    greater_ord(C,S,[],T,[]).

greatereq_ord(C,S,T):-
    g_assign(freshordcounter,0),
    greatereq_ord(C,S,[],T,[]).

greater_ord(_,S,_,_,_):-var(S),!,fail.
greater_ord(C,S,PiS,T,PiT):-var(T),!,checkVar(C,S,PiS,T,PiT).
greater_ord(C,perm(LS,S),PiS,T,PiT):-!,
    append(PiS,LS,EPiS),
    greater_ord(C,S,EPiS,T,PiT),!.
greater_ord(C,S,PiS,perm(LT,T),PiT):-!,
    append(PiT,LT,EPiT),
    greater_ord(C,S,PiS,T,EPiT),!.
greater_ord(_,S,PiS,_,_):-
    S=..[A],
    applyPermAtom(PiS,A,B),
    isFreshOrd(B),!,fail.
greater_ord(_,_,_,T,PiT):-
    T=..[A],
    applyPermAtom(PiT,A,B),
    isFreshOrd(B),!.
greater_ord(C,abs(A,S),PiS,abs(B,T),PiT):-
    applyPermAtom(PiS,A,D),
    applyPermAtom(PiT,B,D),!,
    greater_ord(C,S,PiS,T,PiT),!.
greater_ord(C,abs(A,S),PiS,abs(B,T),PiT):-!,
    freshordname(N),
    applyPermAtom(PiS,A,D),
    applyPermAtom(PiT,B,D1),!,
    greater_ord(C,S,[(D,N)|PiS],T,[(D1,N)|PiT]),!.
greater_ord(C,abs(A,S),PiS,T,PiT):-!,
    freshordname(N),
    applyPermAtom(PiS,A,D),
    greatereq_ord(C,S,[(D,N)|PiS],T,PiT),!.
greater_ord(C,S,PiS,abs(A,T),PiT):-!,
    freshordname(N),
    applyPermAtom(PiT,A,D),
    greater_ord(C,S,PiS,T,[(D,N)|PiT]),!.
greater_ord(C,S,PiS,T,PiT):-
    S=..[F|_],
    T=..[G|ArgsT],
    grprec(F,G),!,
    greater_ord_mul(C,[S],PiS,ArgsT,PiT),!.
greater_ord(C,S,PiS,T,PiT):-
    S=..[F|ArgsS],
    T=..[G|ArgsT],
    (eqprec(F,G);F=G),!,
    (status(F,Stat);Stat=mul),!,
    greater_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,Stat),!.
greater_ord(C,S,PiS,T,PiT):-
    S=..[_|Args],
    member(Si,Args),
    greatereq_ord(C,Si,PiS,T,PiT),!.

greatereq_ord(C,S,PiS,T,PiT):-var(T),!,checkVar(C,S,PiS,T,PiT).
greatereq_ord(C,S,PiS,perm(LT,T),PiT):-!,
    append(PiT,LT,EPiT),
    greatereq_ord(C,S,PiS,T,EPiT),!.
greatereq_ord(_,S,_,_,_):-var(S),!,fail.
greatereq_ord(C,perm(LS,S),PiS,T,PiT):-!,
    append(PiS,LS,EPiS),
    greatereq_ord(C,S,EPiS,T,PiT),!.
greatereq_ord(_,_,_,T,PiT):-
    T=..[A],
    applyPermAtom(PiT,A,B),
    isFreshOrd(B),!.
greatereq_ord(C,abs(A,S),PiS,abs(B,T),PiT):-
    applyPermAtom(PiS,A,D),
    applyPermAtom(PiT,B,D),
    greatereq_ord(C,S,PiS,T,PiT),!.
greatereq_ord(C,abs(A,S),PiS,abs(B,T),PiT):-!,
    freshordname(N),
    applyPermAtom(PiS,A,D),
    applyPermAtom(PiT,B,D1),
    greatereq_ord(C,S,[(D,N)|PiS],T,[(D1,N)|PiT]),!.
greatereq_ord(C,abs(A,S),PiS,T,PiT):-!,
    freshordname(N),
    applyPermAtom(PiS,A,D),
    greatereq_ord(C,S,[(D,N)|PiS],T,PiT),!.
greatereq_ord(C,S,PiS,abs(A,T),PiT):-!,
    freshordname(N),
    applyPermAtom(PiT,A,D),
    greater_ord(C,S,PiS,T,[(D,N)|PiT]),!.
greatereq_ord(C,S,PiS,T,PiT):-
    S=..[F|_],
    T=..[G|ArgsT],
    grprec(F,G),!,
    greater_ord_mul(C,[S],PiS,ArgsT,PiT),!.
greatereq_ord(C,S,PiS,T,PiT):-
    S=..[F|ArgsS],
    T=..[G|ArgsT],
    (eqprec(F,G);F=G),!,
    (status(F,Stat);Stat=mul),!,
    greatereq_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,Stat),!.
greatereq_ord(C,S,PiS,T,PiT):-
    S=..[_|Args],
    member(Si,Args),
    greatereq_ord(C,Si,PiS,T,PiT),!.

greater_ord_stat(C,_,ArgsS,PiS,ArgsT,PiT,mul):-
    greater_ord_mul(C,ArgsS,PiS,ArgsT,PiT),!.
greater_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,lex-lr):-
    greater_ord_lex(C,S,ArgsS,PiS,ArgsT,PiT),!.
greater_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,lex-rl):-
    reverse(ArgsS,RArgsS),
    reverse(ArgsT,RArgsT),
    greater_ord_lex(C,S,RArgsS,PiS,RArgsT,PiT),!.

greatereq_ord_stat(C,_,ArgsS,PiS,ArgsT,PiT,mul):-
    greatereq_ord_mul(C,ArgsS,PiS,ArgsT,PiT),!.
greatereq_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,lex-lr):-
    greatereq_ord_lex(C,S,ArgsS,PiS,ArgsT,PiT),!.
greatereq_ord_stat(C,S,ArgsS,PiS,ArgsT,PiT,lex-rl):-
    reverse(ArgsS,RArgsS),
    reverse(ArgsT,RArgsT),
    greatereq_ord_lex(C,S,RArgsS,PiS,RArgsT,PiT),!.

greater_ord_mul(C,[S1|ArgsS],PiS,ArgsT,PiT):- 
    delete1(Ti,ArgsT,ArgsTi),
    greater_ord(C,S1,PiS,Ti,PiT),!,
    greatereq_ord_mul(C,[S1|ArgsS],PiS,ArgsTi,PiT).
greater_ord_mul(C,[S1|ArgsS],PiS,ArgsT,PiT):- 
    delete1(Ti,ArgsT,ArgsTi),
    greatereq_ord(C,S1,PiS,Ti,PiT),
    greater_ord_mul(C,ArgsS,PiS,ArgsTi,PiT),!.
greater_ord_mul(C,[_|ArgsS],PiS,ArgsT,PiT):- 
    greatereq_ord_mul(C,ArgsS,PiS,ArgsT,PiT).

greatereq_ord_mul(_,[],_,[],_).
greatereq_ord_mul(C,[S1|ArgsS],PiS,ArgsT,PiT):- 
    delete1(Ti,ArgsT,ArgsTi),
    greater_ord(C,S1,PiS,Ti,PiT),!,
    greatereq_ord_mul(C,[S1|ArgsS],PiS,ArgsTi,PiT).
greatereq_ord_mul(C,[S1|ArgsS],PiS,ArgsT,PiT):- 
    delete1(Ti,ArgsT,ArgsTi),
    greatereq_ord(C,S1,PiS,Ti,PiT),
    greatereq_ord_mul(C,ArgsS,PiS,ArgsTi,PiT),!.
greatereq_ord_mul(C,[_|ArgsS],PiS,ArgsT,PiT):- 
    greatereq_ord_mul(C,ArgsS,PiS,ArgsT,PiT).

greater_ord_lex(_,_,[_|_],_,[],_).
greater_ord_lex(C,S,[S1|_],PiS,[T1|ArgsT],PiT):-
    greater_ord(C,S1,PiS,T1,PiT),!,
    greater_ord_mul(C,[S],PiS,ArgsT,PiT),!.
greater_ord_lex(C,S,[S1|ArgsS],PiS,[T1|ArgsT],PiT):-
    greatereq_ord(C,S1,PiS,T1,PiT),!,
    greater_ord_lex(C,S,ArgsS,PiS,ArgsT,PiT).

greatereq_ord_lex(_,_,_,_,[],_).
greatereq_ord_lex(C,S,[S1|_],PiS,[T1|ArgsT],PiT):-
    greater_ord(C,S1,PiS,T1,PiT),!,
    greater_ord_mul(C,[S],PiS,ArgsT,PiT),!.
greatereq_ord_lex(C,S,[S1|ArgsS],PiS,[T1|ArgsT],PiT):-
    greatereq_ord(C,S1,PiS,T1,PiT),!,
    greatereq_ord_lex(C,S,ArgsS,PiS,ArgsT,PiT).
    
%match T1 S1
%apply subst right and check is more particular than T2


%**********(L,R,ML,MR,NL,NR):- L & R terms to be simplified 
%                              ML & MR are the sides of original maximal premise
%                              NL & NR are the simplified side

simplify(C,T1,T2,NT1,NT2):-
    g_assign(freshmatchmax,0),
    simplify(C,T1,NT1),
    simplify(C,T2,NT2),!.

simplify(FC,A,A1):- 
    (rule(FCRuleo,Lo,Ro);
     ruleold(FCRuleo,Lo,Ro);
     rulecur(FCRuleo,Lo,Ro)),
    refreshmatch(FCRuleo,Lo,Ro,FCRule,L,R),
    subtnovar(A,S,C,X),
    matchrule([],L,[],S,MC),
%    write(matched(L,S,MC)),nl,
    append(FCRule,MC,FCRule1),
    simplifyContext(FCRule1,FCRule2),
    removefresh(FCRule2,FCRule3),
    chkfresh(FC,FCRule3),
%    write('***normal rule '), write(rule(FCRuleo,Lo,Ro)),nl,
%    write('***fresh  rule '), write(rule(FCRule,L,R)),nl,
%    write('***bef'), write(C),nl,
    X=R,
%    write('***'), write(C),nl,
    downPerm(C,C1),
%    write('***'), write(C1),nl,
    simplifyPerms(C1,C2),
%    write('***'), write(C2),nl,
    g_read(freshmatchcounter,N),
    g_assign(freshmatchmax,N),
    simplify(FC,C2,A1),!.
simplify(_,A,A).

removefresh([],[]).
removefresh([(A,_)|FC],FC1):-
    isFreshMatch(A),!,
    removefresh(FC,FC1).
removefresh([(A,X)|FC],[(A,X)|FC1]):-
    removefresh(FC,FC1).

applysubs([]).
applysubs([[T,T]|L]):-
    applysubs(L).

isFreshOrd(Name):-
    name(Name,[35,35|_]).

isFreshMatch(Name):-
    name(Name,[35,109|_]).

freshordname(Name):-
    g_read(freshordcounter,N),
    N1 is N+1,
    g_assign(freshordcounter,N1),
    name(N1,L),
    name(Name,[35,35|L]).

freshmatchname(Name):-
    g_read(freshmatchcounter,N),
    N1 is N+1,
    g_assign(freshmatchcounter,N1),
    name(N1,L),
    name(Name,[35,109|L]).

freshname(Name):-
    g_read(freshcounter,N),
    N1 is N+1,
    g_assign(freshcounter,N1),
    name(N1,L),
    name(Name,[35|L]).

max(N,M,N):-N>M,!.
max(_,M,M).

refresh0(C1,T1,T2,C2,T3,T4):-
    g_assign(freshcounter,0),
    refreshone([],T1,L1,T3),
    refreshone(L1,T2,L2,T4),
    refreshContext(C1,L2,C2),
    g_read(freshcounter,N),
    g_read(freshmax,M),
    max(N,M,M1),
    g_assign(freshmax,M1),!.

refresh(C1,T1,T2,C2,T3,T4):-
    g_read(freshmax,N),
    g_assign(freshcounter,N),
    refreshone([],T1,L1,T3),
    refreshone(L1,T2,L2,T4),
    refreshContext(C1,L2,C2),!.

refreshmatch(C1,T1,T2,C2,T3,T4):-
    g_read(freshmatchmax,N),
    g_assign(freshmatchcounter,N),
    refreshonematch([],T1,L1,T3),
    refreshonematch(L1,T2,L2,T4),
    refreshContext(C1,L2,C2),!.

refreshone(L,X,L,X):-var(X),!.
refreshone(L,A,L,Name):-
    A=..[_],
    member2(A,L,Name),!.
refreshone(L,abs(A,T),L2,abs(Name,T1)):-
    member2(A,L,Name),!,
    refreshone(L,T,L2,T1),!.
refreshone(L,abs(A,T),L2,abs(Name,T1)):-!,
    freshname(Name),
    refreshone([[A,Name]|L],T,L2,T1),!.
refreshone(L,perm(LP,T1),L1,perm(LP1,T2)):-
    refreshonelist(L,LP,L,LP1),
    refreshone(L,T1,L1,T2).
refreshone(L,T1,L1,T2):-
    T1=..[F|Args],
    refreshonelist(L,Args,L1,Args1),
    T2=..[F|Args1].

refreshonelist(L,[],L,[]).
refreshonelist(L,[T1|Args1],L2,[T2|Args2]):-
    refreshone(L,T1,L1,T2),
    refreshonelist(L1,Args1,L2,Args2),!.

refreshonematch(L,X,L,X):-var(X),!.
refreshonematch(L,A,L,Name):-
    A=..[_],
    member2(A,L,Name),!.
refreshonematch(L,abs(A,T),L2,abs(Name,T1)):-
    member2(A,L,Name),!,
    refreshonematch(L,T,L2,T1),!.
refreshonematch(L,abs(A,T),L2,abs(Name,T1)):-!,
    freshmatchname(Name),
    refreshonematch([[A,Name]|L],T,L2,T1),!.
refreshonematch(L,perm(LP,T1),L1,perm(LP1,T2)):-
    refreshonematchlist(L,LP,L,LP1),
    refreshonematch(L,T1,L1,T2).
refreshonematch(L,T1,L1,T2):-
    T1=..[F|Args],
    refreshonematchlist(L,Args,L1,Args1),
    T2=..[F|Args1].

refreshonematchlist(L,[],L,[]).
refreshonematchlist(L,[T1|Args1],L2,[T2|Args2]):-
    refreshonematch(L,T1,L1,T2),
    refreshonematchlist(L1,Args1,L2,Args2),!.


refreshContext([],_,[]):-!.
refreshContext([(A,X)|C],L,[(N,X)|C1]):-
    member2(A,L,N),!,
    refreshContext(C,L,C1),!.
refreshContext([(A,X)|C],L,[(A,X)|C1]):-
    refreshContext(C,L,C1),!.

chkfresh(_,[]).
chkfresh(C,[(A,T)|R]):-
    chkcondition(C,A,T),!,
    chkfresh(C,R).

chkcondition([(A,T1)|_],A,T2):-
    T1==T2,!.
chkcondition([_|L],A,T2):-
    chkcondition(L,A,T2).

issubterm(T1,T2):-var(T2),!,T1==T2.
issubterm(T1,T2):-T1==T2,!.
issubterm(T1,T2):-
    T2=..[_|Args],
    issubtermlist(T1,Args).
    
issubtermlist(T1,[T2|_]):-
    issubterm(T1,T2),!.
issubtermlist(T1,[_|Args]):-
    issubtermlist(T1,Args).

subtnovar(T3,_,_,_):-var(T3),!,fail.
subtnovar(T3,T3,X,X).
subtnovar(T3,S,C,X):-
    T3=..[F|Args],
    subtnovarlist(Args,S,ArgsC,X),
    C=..[F|ArgsC].
    
subtnovarlist([T|Args],S,[CT|Args],X):-
    subtnovar(T,S,CT,X).
subtnovarlist([T|Args],S,[T|ArgsC],X):-
    subtnovarlist(Args,S,ArgsC,X).

renameContext([],_,_,[]).
renameContext([(A,T)|FC],A,B,[(B,T)|FC1]):-!,
    renameContext(FC,A,B,FC1),!.
renameContext([(C,T)|FC],A,B,[(C,T1)|FC1]):-!,
    renameTerm(T,A,B,T1),
    renameContext(FC,A,B,FC1),!.

renameTerm(X,_,_,X):-var(X),!.
renameTerm(A,A,B,B):-!.
renameTerm(T,A,B,T1):-
    T=..[F|Args],
    renameTermList(Args,A,B,Args1),
    T1=..[F|Args1],!.

renameTermList([],_,_,[]).
renameTermList([T|Args],A,B,[T1|Args1]):-
    renameTerm(T,A,B,T1),
    renameTermList(Args,A,B,Args1),!.

matchrule(PermS,S,PermT,T,UC):-var(S),var(T),S==T,!,
    difPi(PermS,PermT,S,UC).
matchrule(PermS,S,PermT,T,[]):-var(S),var(T),!,
    applyPermut(PermT,T,T1),
    reverse(PermS,RPermS),
    applyPermut(RPermS,T1,T2),
    S=T2,!.
matchrule(PermS,S,PermT,perm(LP,T),UC):-var(S),!,
    append(PermT,LP,PermT1),
    matchrule(PermS,S,PermT1,T,UC),!.
matchrule(PermS,S,PermT,T,[]):-var(S),!,
    applyPermut(PermT,T,T1),
    reverse(PermS,RPermS),
    applyPermut(RPermS,T1,T2),
    S=T2,!.
matchrule(PermS,perm(LP,S),PermT,T,UC):-var(T),!,
    append(PermS,LP,PermS1),
    matchrule(PermS1,S,PermT,T,UC),!.
matchrule(_,_,_,T,_):-var(T),!,fail.
matchrule(PermS,S,PermT,T,[]):-
    S=..[A], 
    T=..[B],!,
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,A1),!.
matchrule(PermS,abs(A,S),PermT,abs(B,T),[(A1,T)|UC]):-
    applyPermAtom(PermS,A,A1),
    applyPermAtom(PermT,B,B1),
    (A1==B1 -> matchrule(PermS,S,PermT,T,UC)
     ; 
     matchrule([(B1,A1)|PermS],S,PermT,T,UC)),!.
matchrule(PermS,perm(LP,S),PermT,T,UC):-!,
    append(PermS,LP,PermS1),
    matchrule(PermS1,S,PermT,T,UC),!.
matchrule(PermS,S,PermT,perm(LP,T),UC):-!,
    append(PermT,LP,PermT1),
    matchrule(PermS,S,PermT1,T,UC),!.
matchrule(PermS,S,PermT,T,UC):-
    S=..[F|ArgsS], 
    T=..[F|ArgsT],!,
    matchrulelist(PermS,ArgsS,PermT,ArgsT,UC),!.

matchrulelist(_,[],_,[],[]).
matchrulelist(PermS,[S1|ArgsS],PermT,[T1|ArgsT],UC):-
    matchrule(PermS,S1,PermT,T1,UC1),!,
    matchrulelist(PermS,ArgsS,PermT,ArgsT,UC2),
    append(UC1,UC2,UC),!.

compatible(X,T,[],[[X,T]]):-!.
compatible(X,T,[[X1,T1]|L],[[X1,T1]|L]):-
	X==X1,!,T==T1,!.
compatible(X,T,[P|L],[P|L1]):-
	compatible(X,T,L,L1),!.

member1(X,[Y|_]):- X==Y,!.
member1(X,[_|L]):- member1(X,L),!.

member2(X,[[Y,Z]|_],Z):- X==Y,!.
member2(X,[_|L],Z):- member2(X,L,Z),!.

member22(X,[(Y,Z)|_],Z):- X==Y.
member22(X,[_|L],Z):- member22(X,L,Z),!.

delete1(X,[X|L],L).
delete1(X,[Y|L],[Y|Z]):- delete1(X,L,Z).

delete2(X,[Y|L],L):-X==Y,!.
delete2(X,[Y|L],[Y|Z]):- delete2(X,L,Z),!.

%***************************************************************

% C,L,R have fresh names.

simplifyback(C,L):-
    retract(rule(Co,Lo,Ro)),
    (isReducible(Co,Lo,Ro,C,L) ->
     (write('   The rule  '),nl,
      write('    '), write_rule(Co,Lo,Ro),nl,
      write('   can be simplified'),nl,
      simplify(Co,Lo,Ro,Ln,Rn),
      (
       (checkresult(Co,Lo,Ro,Ln,Rn),
	compare(Co,Ln,Rn,OLn,ORn,RED),
	(RED=0 -> (write('unorientable'),nl,!,abort)
	 ; 
	 (
	  simplifyContextTerm(Co,pair(OLn,ORn),CF),
	  refresh0(CF,OLn,ORn,CF0,OLn0,ORn0),
	  write('   The rule  '),nl,
	  write('    '), write_rule(CF0,OLn0,ORn0),nl,
	  write('   has been added'),nl,
	  assertz(rule(CF0,OLn0,ORn0))
	 )
	)
       )
       ; true
      )
     )
     ; assertz(rule(Co,Lo,Ro))
    ),!.
simplifyback(C,L):-
    retract(ruleold(Co,Lo,Ro)),
    (isReducible(Co,Lo,Ro,C,L) ->
     (write('   The rule  '),nl,
      write('    '), write_rule(Co,Lo,Ro),nl,
      write('   can be simplified'),nl,
      simplify(Co,Lo,Ro,Ln,Rn),
      (
       (checkresult(Co,Lo,Ro,Ln,Rn),
	compare(Co,Ln,Rn,OLn,ORn,RED),
	(RED=0 -> (write('unorientable'),nl,!,abort)
	 ; 
	 (
	  simplifyContextTerm(Co,pair(OLn,ORn),CF),
	  refresh0(CF,OLn,ORn,CF0,OLn0,ORn0),
	  write('   The rule  '),nl,
	  write('    '), write_rule(CF0,OLn0,ORn0),nl,
	  write('   has been added'),nl,
	  assertz(rule(CF0,OLn0,ORn0))
	 )
	)
       )
       ; true
      )
     )
     ; assertz(ruleold(Co,Lo,Ro))
    ),!.
simplifyback(_,_).


isReducible(Co,Lo,Ro,C,L):-
    \+ ( \+ (isReducible1(Co,Lo,Ro,C,L))).

isReducible1(Co,Lo,Ro,C,L):-
    (subtnovar(Lo,S,_,_);subtnovar(Ro,S,_,_)),
    matchrule([],L,[],S,MC),
    append(C,MC,C1),
    simplifyContext(C1,C2),
    removefresh(C2,C3),
    chkfresh(Co,C3),!.
